//**************************************************************************/
// Copyright (c) 1998-2008 Autodesk, Inc.
// All rights reserved.
// 
//  Use of this software is subject to the terms of the Autodesk license 
//  agreement provided at the time of installation or download, or which 
//  otherwise accompanies this software in either electronic or hard copy form.
//**************************************************************************/
// DESCRIPTION: Interface for Biped for Max 12
// AUTHOR: Susan Amkraut - created August 28, 2008
//***************************************************************************/

#pragma once

#include "biped11Api.h"


/*! \defgroup BipedInterface12 Biped Interface 12
Max 12 Biped export interface
*/

//! \brief Interface into the biped driver controller.
/*! This contains functions supporting new biped features.
*/
class IBipDriver12 : public IBipDriver11
{
	public:

	/************ Access to the Biped Interface ***************/
	/*! \name Biped Interface Access
	*/
//!@{
		//! \deprecated This has been deprecated as of 3ds Max 2022, please use I_BIPDRIVER12 instead [insensitive language]
		MAX_DEPRECATED static const ULONG I_BIPMASTER12 = 0x9172;    //!< Biped interface identifier
		static constexpr ULONG I_BIPDRIVER12 = 0x9172;    //!< Biped interface identifier

		//! \deprecated This has been deprecated as of 3ds Max 2022, please use GetBipDriver12Interface instead [insensitive language]
		MAX_DEPRECATED BIPExport static IBipDriver12* GetBipMaster12Interface(Animatable& anim);
		//! \brief Gets the Biped interface
		//! \param[in]  anim - the Animatable whose Biped interface is returned
		//! \return the Biped interface
		BIPExport static IBipDriver12* GetBipDriver12Interface(Animatable& anim);
//!@}*/

	/************ Knuckles ***************/
	/*! \name Knuckles
	When Knuckles is set to true, the hand is shrunk, the fingers have 4 links, splay out from the hand, and ThumbMinusOne is set to true.
	When Knuckles is set to false, the fingers and hands appear as in 3ds Max 2009, and ThumbMinusOne is set to false.
	*/
//!@{		
		//! \brief Queries the value of Knuckles
		//! \return true if Knuckles is on, false if not
		virtual BOOL	GetKnuckles() const = 0;

		//! \brief Sets Knuckles
		//! \pre must be in figure mode
		//! \pre ForeFeet must be false
		//! \param[in]  bEnabled - sets Knuckles to the value of bEnabled
		virtual void	SetKnuckles(BOOL bEnabled)=0;
//!@}

		
	/************ ShortThumb  ***************/
	/*! \name ShortThumb
	If ShortThumb is true, the thumb has one less link than the other fingers, except when there is only one link per finger.
	If ShortThumb is false, the thumb has the same number of links as the fingers.
	*/
//!@{		
		//! \brief Queries the value of ShortThumb
		//! \return true if ShortThumb is on, false if not
		virtual BOOL	GetShortThumb() const = 0;

		//! \brief Sets ShortThumb
		//! \pre must be in figure mode
		//! \pre Knuckles must be true
		//! \param[in]  bEnabled - sets ShortThumb to the value of bEnabled
		virtual void	SetShortThumb(BOOL bEnabled)=0;
//!@}

		
	/************ Layer Retargeting  ***************/
	/*! \name Layer Retargeting
	The following functions are used for retargeting biped layer animation, either to the base layer or to a reference biped.
	*/
//!@{
		//! \brief Gets the layer retarget state for a limb
		//! \param[in]  keytrack - KEY_LARM or KEY_RARM or KEY_LLEG or KEY_RLEG
		//! \return true if the limb's retarget state is on, false if not
		virtual bool	GetLimbRetargetState(int keytrack)=0;

		//! \brief Sets the layer retarget state for a limb
		//! \param[in]  keytrack - KEY_LARM or KEY_RARM or KEY_LLEG or KEY_RLEG
		//! \param[in]  bEnabled - sets the limb retarget state to the value of bEnabled.
		//! \see RetargetToBaseLayer(bool IK_Only)
		//! \see RetargetToReferenceBiped(bool IK_Only)
		virtual void	SetLimbRetargetState(int keytrack, bool bEnabled)=0;

		//! \brief Gets the retarget reference biped
		//! \return the retarget reference biped if there is one, NULL if not
		virtual INode	*GetRetargetReferenceBiped()=0;
		
		//! \brief Sets the retarget reference biped
		//! \param[in]  node - sets the reference biped to the node, which must be any node from a different biped.
		//! \see RetargetToReferenceBiped(bool IK_Only)
		//! \return true if the reference biped was set correctly, false if not
		virtual bool	SetRetargetReferenceBiped(INode *node)=0;

		//! \brief Retargets the current biped layer to its base layer.
		/*! For every limb whose retarget state is active, calculates the hand/foot position at every key,
		maintaining IK constraints to the base layer.
		\pre A layer must be active
		\see SetLayerActive(int index, bool onOff)
		\param[in]  IK_Only - When IK_Only is on, the biped's constrained hands and feet are retargeted only during the frames on which they are IK controlled.
		When IK_Only is off, the hands and feet are retargeted during both IK and FK keys.
		*/
		virtual void	RetargetToBaseLayer(bool IK_Only)=0;

		//! \brief Retargets the current biped layer to the reference biped.
		/*! For every limb whose retarget state is active, calculates the hand/foot position at every key,
		maintaining IK constraints to the retarget biped.
		\pre A retarget biped must be specified.
		\see SetRetargetReferenceBiped()
		\param[in]  IK_Only - When IK_Only is on, the biped's constrained hands and feet are retargeted only during the frames on which they are IK controlled.
		When IK_Only is off, the hands and feet are retargeted during both IK and FK keys.
		*/
		virtual void	RetargetToReferenceBiped(bool IK_Only)=0;
//!@}
};

//! \deprecated This has been deprecated as of 3ds Max 2022, please use IBipDriver12 instead [insensitive language]
typedef MAX_DEPRECATED IBipDriver12 IBipMaster12;
