#pragma once
//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
// Use of this software is subject to the terms of the Autodesk license
// agreement provided at the time of installation or download, or which
// otherwise accompanies this software in either electronic or hard copy form.
//**************************************************************************/

#include <MaxUsd/MaxUSDAPI.h>

#include <pxr/base/tf/declarePtrs.h>
#include <pxr/base/tf/registryManager.h>
#include <pxr/base/tf/singleton.h>
#include <pxr/base/tf/token.h>
#include <pxr/base/tf/weakBase.h>
#include <pxr/base/vt/dictionary.h>
#include <pxr/pxr.h>

#include <string>

PXR_NAMESPACE_OPEN_SCOPE

TF_DECLARE_WEAK_PTRS(MaxUsdJobContextRegistry);

/// We provide macros that are entry points into the job context logic.
class MaxUsdJobContextRegistry : public TfWeakBase
{
public:
	/// An job context basically wraps a function that tweaks the set of import/export options. This
	/// job context has a name and UI components, as well as an enabler function that allows
	/// specifying the options dictionary.
	///
	/// To register an export job context, you need to use the REGISTER_EXPORT_JOB_CONTEXT macro for
	/// each export job context supported by the library.
	///
	/// In order for the core system to discover the plugin, you need a \c plugInfo.json that
	/// declares job contexts. \code
	/// {
	///   "Plugins": [
	///     {
	///       "Info": {
	///         "MaxUsd": {
	///          "JobContextPlugin": {
	///          }
	///         }
	///       },
	///       "Name": "myUsdPlugin",
	///       "LibraryPath": "../myUsdPlugin.dll",
	///       "Type": "library"
	///     }
	///   ]
	/// }
	/// \endcode

	/// Enabler function, returns a dictionary containing all the options for the context.
	typedef std::function<VtDictionary()> EnablerFn;

	/// Get all registered export job contexts:
	static TfTokenVector ListJobContexts()
	{
		return GetInstance()._ListJobContexts();
	}

	/// All the information registered for a specific job context.
	struct ContextInfo
	{
		TfToken jobContext;
		TfToken niceName;
		TfToken exportDescription;
		EnablerFn exportEnablerCallback;
		TfToken importDescription;
		EnablerFn importEnablerCallback;

		ContextInfo() = default;

		ContextInfo(
			const TfToken& jc,
			const TfToken& nn,
			const TfToken& edsc,
			EnablerFn eef,
			const TfToken& idsc,
			EnablerFn ief)
				: jobContext(jc)
				, niceName(nn)
				, exportDescription(edsc)
				, exportEnablerCallback(eef)
				, importDescription(idsc)
				, importEnablerCallback(ief)
		{
		}
	};

	/// Gets the conversion information associated with \p jobContext on export and import
	static const ContextInfo& GetJobContextInfo(const TfToken& jobContext)
	{
		return GetInstance()._GetJobContextInfo(jobContext);
	}

	/// Registers an export job context, with nice name, description and enabler function.
	///
	/// \param jobContext name will be used directly as one of the valid values of the job context option.
	/// \param niceName is the name displayed in the options dialog.
	/// \param description is displayed as a tooltip in the  options dialog.
	/// \param enablerFct will be called after option parsing to enable context specific options.
	MaxUSDAPI void RegisterExportJobContext(const std::string& jobContext, const std::string& niceName,
			const std::string& description, EnablerFn enablerFct, bool fromPython = false);

	/// Registers an import job context, with nice name, description and enabler function.
	///
	/// \param jobContext name will be used directly as one of the valid values of the job context option.
	/// \param niceName is the name displayed in the options dialog.
	/// \param description is displayed as a tooltip in the  options dialog.
	/// \param enablerFct will be called after option parsing to enable context specific options.
	MaxUSDAPI void RegisterImportJobContext(const std::string& jobContext, const std::string& niceName,
			const std::string& description, EnablerFn enablerFct, bool fromPython = false);

	MaxUSDAPI static MaxUsdJobContextRegistry& GetInstance();

private:
	MaxUSDAPI TfTokenVector _ListJobContexts();
	MaxUSDAPI const ContextInfo& _GetJobContextInfo(const TfToken&);

	MaxUsdJobContextRegistry();
	~MaxUsdJobContextRegistry();
	friend class TfSingleton<MaxUsdJobContextRegistry>;
};

#define REGISTER_EXPORT_JOB_CONTEXT(name, niceName, description, enablerFct)                                       \
	TF_REGISTRY_FUNCTION(MaxUsdJobContextRegistry)                                                                 \
	{                                                                                                              \
		MaxUsdJobContextRegistry::GetInstance().RegisterExportJobContext(name, niceName, description, enablerFct); \
	}

#define REGISTER_EXPORT_JOB_CONTEXT_FCT(name, niceName, description)             \
	static VtDictionary _ExportJobContextEnabler_##name();                       \
	TF_REGISTRY_FUNCTION(MaxUsdJobContextRegistry)                               \
	{                                                                            \
		MaxUsdJobContextRegistry::GetInstance().RegisterExportJobContext(        \
				#name, niceName, description, &_ExportJobContextEnabler_##name); \
	}                                                                            \
	VtDictionary _ExportJobContextEnabler_##name()

PXR_NAMESPACE_CLOSE_SCOPE
