#pragma once

//**************************************************************************/
// Copyright (c) 1998-2019 Autodesk, Inc.
// All rights reserved.
//
// These coded instructions, statements, and computer programs contain
// unpublished proprietary information written by Autodesk, Inc., and are
// protected by Federal copyright law. They may not be disclosed to third
// parties or copied or duplicated in any form, in whole or in part, without
// the prior written consent of Autodesk, Inc.
//**************************************************************************/
// DESCRIPTION: Base class for builder options
// AUTHOR: Autodesk Inc.
//***************************************************************************/

#include <pxr/usd/sdf/path.h>
#include <pxr/base/vt/dictionary.h>

#include "MaxUsd/Utilities/Logging.h"
#include "USDIOCallback.h"

namespace MAXUSD_NS_DEF {

// \brief Merges all the jobContext arguments dictionaries found while exploring the jobContexts into a
// single one. Also checks for conflicts and errors.
//
// \param[in] isExport if we are calling the import or the export jobContext callback.
// \param[in] contexts jobContexts to merge.
//
// \param[out] allContextArgs dictionary of all extra jobContext arguments merged together.
// \return true if the merge was successful, false if a conflict or an error was detected.
MaxUSDAPI bool MergeJobContexts(bool isExport, const std::set<std::string>& contexts, pxr::VtDictionary& allContextArgs);

/**
* \brief Class for getting and setting builder options.
*/
class SceneBuilderOptions
{
public:
	MaxUSDAPI virtual ~SceneBuilderOptions() = default;
	/**
	 * \brief Returns the builder's logging options.
	 * \return The logging options.
	 */
	MaxUSDAPI Log::Options& GetLogOptions();
	// Const version
	MaxUSDAPI const Log::Options& GetLogOptions() const;

	/**
	 * \brief Gets all registered callbacks for the given event type.
	 * \param type The event type for which to get the callbacks.
	 * \param eventCallbacks Callback vector to be filled.
	 */
	MaxUSDAPI void GetCallBacks(USDIOEventType type, std::vector<std::shared_ptr<USDIOCallback>>& eventCallbacks) const;

	/**
	 * \brief Returns all registered callbacks organized by event types and callback Id.
	 * \return A map of all registered callbacks.
	 */
	MaxUSDAPI const std::map<USDIOEventType, std::map<std::wstring, std::shared_ptr<USDIOCallback>>>& GetCallBacks()
			const;

	/**
	 * \brief Removes callbacks by event type and/or id.
	 *  - If neither are specified, all callbacks are removed.
	 *  - If only the event type is specified, all callbacks for that event type are removed.
	 *  - If only the callback id is specified, all callbacks with that Id are removed, whatever the event type.
	 *  - If both are specified, only the callback of that Id for that event type is removed.
	 * \param type The event type for which to remove the callback.
	 * \param id The id of the callback to be removed.
	 */
	MaxUSDAPI void RemoveCallbacks(USDIOEventType type = USDIOEventType::None, const std::wstring& id = L"");

	/**
	 * \brief Registers a new callback.
	 * \param type The event type for which to register the callback.
	 * \param id The callback's Id.
	 * \param callback The callback to be added.
	 */
	MaxUSDAPI void AddCallback(USDIOEventType type, const std::wstring& id, std::shared_ptr<USDIOCallback> callback);

protected:
	/// Logging options.
	Log::Options logOptions;
	/// Registered callbacks.
	std::map<USDIOEventType, std::map<std::wstring, std::shared_ptr<USDIOCallback>>> callbacks;
};
}