//**************************************************************************/
// Copyright (c) 2019 Autodesk, Inc.
// All rights reserved.
//
// Use of this software is subject to the terms of the Autodesk license 
// agreement provided at the time of installation or download, or which 
// otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/
// DESCRIPTION: Description of an interface of a proxy holding a USDStage
// AUTHOR: Autodesk Inc.
//**************************************************************************/

#pragma once

#include <ifnpub.h>
#include <ref.h>

#include <pxr/usd/usd/stage.h>

/// ID if the USD I/O operation interface.
static const Interface_ID IUSDStageProvider_ID(0x6be93509, 0x325e2773);

// No way to assign an obviously unique value. Let's hope this one is.
constexpr const auto REFMSG__IUSDSTAGEPROVIDER_MSGBASE = REFMSG_USER + 9560;
constexpr const auto REFMSG_IUSDSTAGEPROVIDER_STAGE_CHANGED = REFMSG__IUSDSTAGEPROVIDER_MSGBASE + 1;

enum
{
	fnIdReload,
	fnIdSetRootLayer,
	fnIdGetUsdPreviewSurfaceMaterials,
	fnIdSetPrimvarChannelMappingDefaults,
	fnIdSetPrimvarChannelMapping,
	fnIdGetPrimvarChannel,
	fnIdIsMappedPrimvar,
	fnIdGetMappedPrimvars,
	fnIdClearMappedPrimvars
};

/**
 * \brief Exposes an interface for proxies able to provide a USDStage
 */
class IUSDStageProvider : public FPMixinInterface
{
 public:
	// Function Publishing System
	// Function Map For Mixin Interface
	//*************************************************
	BEGIN_FUNCTION_MAP
	 VFN_0(fnIdReload, Reload);
	 VFN_2(fnIdSetRootLayer, SetRootLayerMXS, TYPE_STRING, TYPE_STRING);
	 FN_1(fnIdGetUsdPreviewSurfaceMaterials, TYPE_MTL, GetUsdPreviewSurfaceMaterials, TYPE_BOOL);
	 VFN_0(fnIdSetPrimvarChannelMappingDefaults, SetPrimvarChannelMappingDefaults);
	 VFN_2(fnIdSetPrimvarChannelMapping, SetPrimvarChannelMapping, TYPE_STRING, TYPE_VALUE);
	 FN_1(fnIdGetPrimvarChannel, TYPE_VALUE, GetPrimvarChannel, TYPE_STRING);
	 FN_0(fnIdGetMappedPrimvars, TYPE_STRING_TAB_BV, GetMappedPrimvars);
	 FN_1(fnIdIsMappedPrimvar, TYPE_BOOL, IsMappedPrimvar, TYPE_STRING);
	 VFN_0(fnIdClearMappedPrimvars, ClearMappedPrimvars);
	END_FUNCTION_MAP

	/// Return a weak pointer to the stage held by this provider.
	virtual pxr::UsdStageWeakPtr GetUSDStage() = 0;

	/// Reload all layers of the stage held by this provider.
	virtual void Reload() = 0;

	/// Set the root layer and mask of the stage held by this provider.
	virtual void SetRootLayer(const wchar_t* rootLayer, const wchar_t* stageMask) = 0;

	/// Set the root layer and mask of the stage held by this provider. MXS function.
	virtual void SetRootLayerMXS(const wchar_t* rootLayer, const wchar_t* stageMask) = 0;

	// Returns the multi-materials representing the UsdPreviewSurface materials in the stage.
	virtual Mtl* GetUsdPreviewSurfaceMaterials(bool update) = 0;

	// Sets defaults primvar to channels mappings.
	virtual void SetPrimvarChannelMappingDefaults() = 0;

	// Sets a primvar to channel mapping.
	virtual void SetPrimvarChannelMapping(const wchar_t* primvarName, Value* channel) = 0;

	// Gets the channel to which a primvar maps to.
	virtual Value* GetPrimvarChannel(const wchar_t* primvarName) = 0;
	
	// Gets the list of all currently mapped primvars.
	virtual Tab<const wchar_t*> GetMappedPrimvars() const = 0;

	// Checks if a primvar is currently mapped to a channel.
	virtual bool IsMappedPrimvar(const wchar_t* primvarName) = 0;

	// Clears all primvar mappings.
	virtual void ClearMappedPrimvars() = 0;
};