//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
// Use of this software is subject to the terms of the Autodesk license 
// agreement provided at the time of installation or download, or which 
// otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/
#pragma once
#include "MaxUsd/MappedAttributeBuilder.h"

class MaxMeshConversionOptions
{
public:
	
	/**
	 * \brief Conversion mode for normals. Normals can either be exported as primvars, or as simple attributes.
	 * \comment value of enum is important as it reflects the index in the ui of qcombobox
	 */
	enum class MaxUSDAPI NormalsMode
	{
		AsPrimVar = 0,
		AsAttribute = 1,
		None = 2
	};

	/**
	 * \brief Mesh format to be used when converting to USD. Meshes can either export as Polygonal meshes,
	 * Triangular meshes, or use the same format as in the scene (EditMesh -> Triangles, EditPoly -> Polys)
	 */
	enum class MaxUSDAPI MeshFormat
	{
		FromScene = 0,
		PolyMesh = 1,
		TriMesh = 2
	};

	/**
	 * \brief Constructor.
	 */
	MaxUSDAPI MaxMeshConversionOptions();

	/**
	 * \brief Sets default mesh conversion options. This includes a call to SetDefaultChannelPrimvarMappings().
	 */
	MaxUSDAPI void SetDefaults();

	/**
	 * \brief Sets default channel to primvar mappings.
	 *		alpha -> "displayOpacity"
	 *		shading -> "mapShading"
	 *		vertex color -> "displayColor"
	 *		1-N -> mapN
	 */
	MaxUSDAPI void SetDefaultChannelPrimvarMappings();

	/**
	 * \brief Returns the channel to primvar map.
	 * \return The mapping table.
	 */
	MaxUSDAPI const std::map<int, maxUsd::MappedAttributeBuilder::Config>& GetChannelMappings() const;

	/**
	 * \brief Sets the channel to primvar map.
	 * \param mappings The mapping table.
	 */
	MaxUSDAPI void SetChannelMappings(const std::map<int, maxUsd::MappedAttributeBuilder::Config>& mappings);

	/**
	 * \brief Gets the normal conversion mode.
	 * \return The normal conversion mode
	 */
	MaxUSDAPI NormalsMode GetNormalMode() const;

	/**
	 * \brief Sets the normal conversion mode.
	 * \param normalMode The normal mode to be set.
	 */
	MaxUSDAPI void SetNormalsMode(NormalsMode normalMode);

	/**
	 * \brief Gets the MeshFormat to be used.
	 * \return The mesh format.
	 */
	MaxUSDAPI MeshFormat GetMeshFormat() const;

	/**
	 * \brief Sets the mesh format to be used.
	 * \param meshFormat The mesh format to be set.
	 */
	MaxUSDAPI void SetMeshFormat(MeshFormat meshFormat);
	
	/**
	 * \brief  Sets whether or not the Object-offset transform should be baked into the geometry.
	 * Otherwise, an Xform will be used to represent the node transform's versus the offset.
	 * The object-offset transform is the offset of the object from the node it is attached too.
	 * \param bakeObjectOffset True if we should bake the transform, false otherwise.
	 */
	MaxUSDAPI void SetBakeObjectOffsetTransform(bool bakeObjectOffset);

	/**
	 * \brief Gets whether or not the Object-offset transform should be baked into the geometry.
	 * \return True if object offset transform are baked, false otherwise.
	 */
	MaxUSDAPI bool GetBakeObjectOffsetTransform() const;

	/**
	 * \brief  Sets whether or not to preserve max edge orientation.
	 * \param preserve True if we should preserve the edge orientation, false otherwise.
	 */
	MaxUSDAPI void SetPreserveEdgeOrientation(bool preserve);

	/**
	 * \brief Gets whether or not to preserve max edge orientation.
	 * \return True if we should preserve the edge orientation, false otherwise.
	 */
	MaxUSDAPI bool GetPreserveEdgeOrientation() const;

	/**
	 * \brief Configures a channel to primvar mapping. The specified channel will be exported using the given
	 * configuration - essentially a target primvar name and a type.
	 * \param channel The channel to configure.
	 * \param config The configuration to use when exporting the given channel to a primvar. It is possible to
	 * specify we do not want this channel exported by specifying an empty target primvar name.
	 */
	MaxUSDAPI void SetChannelPrimvarConfig(int channel, const maxUsd::MappedAttributeBuilder::Config& config);

	/**
	 * \brief Returns the primvar configuration for the specified channel.
	 * \param channelId The channel 
	 * \return The primvar configuration.
	 */
	MaxUSDAPI const maxUsd::MappedAttributeBuilder::Config& GetChannelPrimvarConfig(int channelId) const;

protected: 
	// Normal options
	NormalsMode normalMode = NormalsMode::AsPrimVar;
	// How meshes should be exported.
	MeshFormat meshFormat = MeshFormat::FromScene;
	// Bake the offset from a node's pivot in the geometry. Otherwise,
	// an additional Xform will be used for the node.
	bool bakeObjectOffsetTransform = true;
	// Preserve edge orientation on export, 
	// When true, we won't try to make the mesh planar, avoiding triangulation of a few faces
	// This will result in an exported mesh that looks like the original
	bool preserveEdgeOrientation = false;
	// Channel primvar output configuration.
	std::map<int, maxUsd::MappedAttributeBuilder::Config> channelToPrimvarConfig;
};