//**************************************************************************/
// Copyright (c) 1998-2020 Autodesk, Inc.
// All rights reserved.
//
// Use of this software is subject to the terms of the Autodesk license 
// agreement provided at the time of installation or download, or which 
// otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/
// DESCRIPTION: Usd Mesh Conversion Options
// AUTHOR: Autodesk Inc.
//***************************************************************************/

#pragma once

#include "MaxUsd/MaxUSDAPI.h"

 /// This class exposes methods for getting and setting primvar/channel mapping options.
 class PrimvarMappingOptions
{
public:
	/**
	 * \brief Constructor.
	 */
	MaxUSDAPI PrimvarMappingOptions();

	/**
	 * \brief Sets defaults primvar to channels mappings.
	 */
	MaxUSDAPI void SetDefaultPrimvarChannelMappings();

	/**
	 * \brief Returns the primvar to channel map.
	 * \return The mapping table.
	 */
	MaxUSDAPI const std::map<std::string, int>& GetPrimvarMappings() const;

	/**
	 * \brief Sets a primvar to channel mapping.
	 * \param primvar The name of the primvar.
	 * \param channel The channel this primvar should be imported to.
	 */
	MaxUSDAPI void SetPrimvarChannelMapping(const std::string& primvar, int channel);

	/**
	 * \brief Gets the channel to which a primvar maps to.
	 * \param primvar The name of the primvar for which to retrieve the target channel.
	 * \return The target channel for this primvar.
	 */
	MaxUSDAPI int GetPrimvarChannelMapping(const std::string& primvar) const;

	/**
	 * \brief Returns the list of all currently mapped primvars.
	 * \param primvars The vector of primvar names to be filled.
	 */
	MaxUSDAPI void GetMappedPrimvars(std::vector<std::wstring>& primvars) const;

	/**
	 * \brief Checks if a primvar is currently mapped to a channel.
	 * \param primvar The primvar to check.
	 * \return True if the primvar is mapped, false otherwise.
	 */
	MaxUSDAPI bool IsMappedPrimvar(const std::string& primvar) const;

	/**
	 * \brief Gets whether or not to import primvars that are not explicitely mapped.
	 * \return True if unmapped primvars should be imported, false otherwise.
	 */
	MaxUSDAPI bool GetImportUnmappedPrimvars() const;

	/**
	 * \brief Sets whether or not to import primvars that are not explicitely mapped. If true, try to
	 * find the most appropriate channels for each unmapped primvar, based on their types.
	 * \param importUnmappedPrimvars If true, unmapped primvars will be imported.
	 */
	MaxUSDAPI void SetImportUnmappedPrimvars(bool importUnmappedPrimvars);

	/**
	 * \brief Clears all primvar mappings.
	 */
	MaxUSDAPI void ClearMappedPrimvars();
 	
	static MaxUSDAPI const int invalidChannel = INT_MIN;
	
protected:
	std::map<std::string, int> primvarToChannnelMappings;
	bool importUnmappedPrimvars = false;
};
