//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
// Use of this software is subject to the terms of the Autodesk license 
// agreement provided at the time of installation or download, or which 
// otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/
#pragma once

#include "MaxUsd/MaxTokens.h"
#include "MaxUsd/Builders/USDSceneBuilderOptions.h"

PXR_NAMESPACE_OPEN_SCOPE

class MaxUsdWriteJobContext;

/// \brief Base class for all built-in and user-defined prim writers. Translates 3dsMax
/// nodes to USD prims.
///
/// The PrimWriter is only responsible for providing translation of the 3dsMax Object referenced by the received Node. It should therefore not attempt to handle instancing, material assignment, and the transform of the Node itself.
/// Instancing is handled by the calling code - if an object is instanced across multiple nodes, the PrimWriter is only called once, on the first node referencing the instanced object. The required Xform prim hierarchy is already generated.
/// Similarly, the Node's transform is applied by the calling code, on the UsdGeomXformable prim built by the PrimWriter, *after* it is run. If the USD prim is *not* a UsdGeomXformable, a warning is raised, but it doesn't prevent the export from continuing.
class MaxUsdPrimWriter
{
public:
	/// \brief The level of support a writer can offer for a given context
	///
	/// A basic writer that gives correct results across most contexts should
	/// report `Fallback`, while a specialized writer that really shines in a
	/// given context should report `Supported` when the context is right and
	/// `Unsupported` if the context is not as expected.
	enum class ContextSupport
	{
		Unsupported,
		Supported,
		Fallback
	};

	/**
	 * \brief Constructor.
	 * \param jobCtx The write job context (options, target file, stage, etc.)
	 */
	MaxUSDAPI MaxUsdPrimWriter(const MaxUsdWriteJobContext& jobCtx);

	/**
	 * \brief Destructor.
	 */
	MaxUSDAPI virtual ~MaxUsdPrimWriter();

	/**
	 * \brief Returns the level of support this writer can provide for a given 3dsMax node.
	 * \param node The node to check for support.
	 * \return The level of support.
	 */
	MaxUSDAPI virtual ContextSupport CanConvert(INode* node) { return ContextSupport::Unsupported; };
	
	/**
	 * \brief The prim type you are writing to.
	 * For performance reasons, all prims get created ahead of time in a single `SdfChangeBlock`. This means
	 * the prim writers are not responsible for creating the prims. The type specified here is mostly a hint
	 * for that first creation pass and, if required, it can be overridden from the `Write()` method (by
	 * defining a prim at the same path with a different type).
	 * Unless you always force the creation of an Xform prim (see `RequireXformPrim()`), you should return
	 * an Xformable type here; otherwise, it is not possible to apply the node's transform onto the prim
	 * later (an error will be raised in this scenario).
	 * \param node The 3dsMax node for which to get the target USD prim type.
	 * \return The prim's type.
	 */
	MaxUSDAPI virtual TfToken GetPrimType(INode* node) { return pxr::MaxUsdPrimTypeTokens->Xform; };

	/**
	 * \brief Responsible for writing the prim's attribute for the given context. This is where the translation from
	 * the 3dsMax node to the USD prim happens.
	 * \param sourceNode The node we are translating to USD.
	 * \param targetPrim The prim we are writing to.
	 * \param applyOffsetTransform Whether or not the object offset transform should be applied onto the prim. Can only
	 * be true if HandlesObjectOffsetTransform() returned true, otherwise it will always be false.
	 * \return True if the write operation was successful, false otherwise.
	 */
	MaxUSDAPI virtual bool Write(INode* sourceNode, UsdPrim& targetPrim, bool applyOffsetTransform) { return false; }

	/**
	 * \brief Returns the name of this prim writer.
	 * \return The writer's name.
	 */
	MaxUSDAPI virtual WStr GetWriterName() { return WStr(); };

	/**
	 * \brief In a few scenarios, you need two prims to properly represent an INode. One for its transform, and one 
	 * for the object it references (for example, in the case of a non-identity object offset transform, it must not 
	 * inherit the transform, so you cannot use a single prim for the INode). When you do need to perform such a 
	 * split, the object's prim has the same name as the node's prim, with an added suffix - the string returned here.
	 * \return A suffix for the object prim.
	 */
	MaxUSDAPI virtual TfToken GetObjectPrimSuffix(INode* /*node*/)
	{
		return TfToken("Object");
	};

	/**
	 * \brief Returns the name that should be used for the prim. The base implementation should be sufficient in most
	 * cases, unless prim writers want to customize the prim's name. If so, it is their responsibility to ensure that
	 * the given name is unique amongst siblings.
	 * \param node The node to get a prim name for.
	 * \param suggestedName A suggested name, from the node's name, and ensured unique amongst siblings. 
	 * \return The name that should be used (will be part of the prim's path /foo/bar/{Prim name}
	 */
	MaxUSDAPI virtual std::string GetPrimName(INode* node, const std::string& suggestedName)
	{
		return suggestedName;
	};

	/**
	 * \brief Returns the requirement to split the object from its transform in the scene.
	 * - ForOffsetObjects : This should be the case for most objects, this means we require an Xform if an object offset
	 * is applied to the object. Indeed, object offset transforms should not be inherited, so we need an Xform prim to
	 * encode the node's transform (the children of the node will export to children of this prim), and another for the
	 * object itself, which will be exported to a prim under that Xform. It will contain the object offset transform, and
	 * it will not have children.
	 * - Always :For cases where we always need to have a seperate prim for the node's transform. For example 
	  if we are exporting to a gprim, which, unless we explicitely allow it, should not be nested. This could also 
	  be the case if we need to add an inherent transform to the object's prim, part of the translation, and we never want that 
	 * transform to be inherited. 
	 * - Never : To be used if we know we never want to split the node's transform from its object. For example we could
	 *  be baking the object offset transform into the geometry itself - in this scenario, we avoid the *need of an extra
	 *  Xform entirely.
	 * \param node The node to test for the xform split requirement.
	 * \return The Xform split requirement for this object.
	 */
	MaxUSDAPI virtual maxUsd::XformSplitRequirement RequiresXformPrim(INode* node)
	{
		return maxUsd::XformSplitRequirement::ForOffsetObjects;
	};

	/**
	 * \brief Whether or not we want to manually handle the object offset transform in the Write(). If we
	 * return true here, then the applyOffsetTransform argument of Write() can sometimes be set to true.
	 * \return True if the prim writer intends to handle the object offset transform, false otherwise.
	 */
	MaxUSDAPI virtual bool HandlesObjectOffsetTransform()
	{
		return false;
	};

	/**
	 * \brief Returns the material assignment requirement for this node. Some prim writers may not want the prims
	 * they produce to be assigned the source node's material automatically.
	 * \param node The 3dsMax node
	 * \return The material assignment requirement.
	 */
	MaxUSDAPI virtual maxUsd::MaterialAssignRequirement RequiresMaterialAssignment(INode* node)
	{
		return maxUsd::MaterialAssignRequirement::Default;
	};

	/**
	 * \brief Returns the instancing requirement for this Prim Writer. Some Prim Writers may want to handle instancing
	 * themselves, or disable it entirely.
	 * \return The instancing requirement. If default, instancing is handled automatically. Otherwise, it is left to the
	 * Prim writer to decide what to do.
	 */
	MaxUSDAPI virtual maxUsd::InstancingRequirement RequiresInstancing()
	{
		return maxUsd::InstancingRequirement::Default;
	};

	MaxUSDAPI const UsdStageRefPtr& GetUsdStage() const;
	MaxUSDAPI const std::string& GetFilename() const;
	MaxUSDAPI virtual boost::python::dict GetNodesToPrims() const;

protected:
	/// Gets the current global export args in effect.
	MaxUSDAPI const USDSceneBuilderOptions& GetExportArgs() const;
	MaxUSDAPI const MaxUsdWriteJobContext& GetJobContext() const;

private:
	const MaxUsdWriteJobContext& writeJobCtx;
};

typedef std::shared_ptr<MaxUsdPrimWriter> MaxUsdPrimWriterSharedPtr;

PXR_NAMESPACE_CLOSE_SCOPE
