//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
// Use of this software is subject to the terms of the Autodesk license 
// agreement provided at the time of installation or download, or which 
// otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/

#pragma once
#include "MaxUsd/Translators/primWriter.h"

#include <pxr/pxr.h>

#include <functional>
#include <string>

PXR_NAMESPACE_OPEN_SCOPE

/**
 * \brief Provides functionality to register and lookup USD writer plugins
 * for 3dsMax nodes.
 *
 * Use PXR_MAXUSD_REGISTER_WRITER(uniqueName, writerClass) to register a writer
 * class with the registry.
 *
 * Prim writers derive from the PrimWriter class and are expected to implement 
 * several methods to specify what objects they are able to support, and how to actually
 * perform the conversion.

 * In order for the core system to discover the plugin at export time, you need a
 * plugInfo.json specifying the plugin's type.
 * \code
 * {
 *   "Plugins":[
 *      {
 *         "Info":{
 *            "MaxUsd":{
 *               "PrimWriter" : {}
 *            }
 *         },
 *         "Name":"myTranslatorPlugin",
 *         "Type":"library",
 *         "LibraryPath":"myTranslatorPlugin.dll"
 *      }
 *   ]
 * }
 * \endcode
 */
struct MaxUsdPrimWriterRegistry
{
    // Writer factory function, i.e. a function that creates a prim writer
    // for the given 3dsMax export job context.
    typedef std::function<MaxUsdPrimWriterSharedPtr(
        const MaxUsdWriteJobContext&)>
        WriterFactoryFn;
    
    /**
     * \brief Register a new prim writer via it's factory function.
     * \param A unique key for the writer. If not unique, an error is returned.
     * \param fn A factory function providing a MaxUsdPrimWriter subclass that can be used to write/
     * \param fromPython  true if the writer is registered from python.
     */
    MaxUSDAPI static void Register(const std::string& key, WriterFactoryFn fn, bool fromPython = false);

    /**
     * \brief Unregisters a prim writer.
     * \param key The unique key for the writer.
     */
    MaxUSDAPI static void Unregister(const std::string& key);
    
    /**
     * \brief Returns all prim writers for a given write job context.
     * \param jobCtx The write job context to get writer for.
     * \return All registered prim writers
     */
    MaxUSDAPI static std::vector<MaxUsdPrimWriterSharedPtr> GetAllWriters(const MaxUsdWriteJobContext& jobCtx);

    /**
     * \brief Returns prim writers which can be used to translate a given object, given from a list of candidate writers.
     * Prioritize ContextSupport::Supported over ContextSupport::Fallback. Other than this, writers are returned in the order
     * they are they are received as candidates.
     * \param node The 3dsMax node we are finding a writer for.
     * \param candidateWriters A vector of candidate prim writers, to be filtered.
     * \return Ordered list of applicable prim writers (Writers with explicit support first, followed with writers that can be
     * used as fallback for this object).
     */
    MaxUSDAPI static std::vector<MaxUsdPrimWriterSharedPtr> FindWriters(
        INode* node, const std::vector<MaxUsdPrimWriterSharedPtr>& candidateWriters);
};

#define PXR_MAXUSD_REGISTER_WRITER(writerClass)                                         \
    TF_REGISTRY_FUNCTION_WITH_TAG(MaxUsdPrimWriterRegistry, writerName##_##writerClass) \
    {                                                                                   \
        MaxUsdPrimWriterRegistry::Register(                                             \
            #writerClass,                                                               \
            [](const MaxUsdWriteJobContext&  jobCtx) {                                  \
                return std::make_shared<writerClass>(jobCtx);                           \
            });                                                                         \
    }

PXR_NAMESPACE_CLOSE_SCOPE