#pragma once
//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
// These coded instructions, statements, and computer programs contain
// unpublished proprietary information written by Autodesk, Inc., and are
// protected by Federal copyright law. They may not be disclosed to third
// parties or copied or duplicated in any form, in whole or in part, without
// the prior written consent of Autodesk, Inc.
//**************************************************************************/

#include "WriteJobContext.h"

#include <pxr/base/tf/refBase.h>
#include <pxr/base/tf/registryManager.h>
#include <pxr/usd/usd/common.h>
#include <pxr/usd/usd/timeCode.h>

class Mtl;

PXR_NAMESPACE_OPEN_SCOPE

/// \brief The ShaderWriter base class from which material writers need to inherit from. A ShaderWriter instance is created for each material needing translation.
///
/// Two methods need to be implemented to have functional ShaderWriter:
/// - `CanExport(exportArgs)` – a static classmethod which returns an enum value stating if the export context is `maxUsd.ShaderWriter.ContextSupport::Supported`, or `Unsupported`, or that the class acts as a Fallback.
/// - `Write()` – the write method called to properly export the material
class MaxUsdShaderWriter
///
{
public:
	MaxUSDAPI MaxUsdShaderWriter(
		Mtl*                   material,
		const SdfPath&         usdPath,
		MaxUsdWriteJobContext& jobCtx);

	virtual ~MaxUsdShaderWriter() =default;

	/// The level of support a writer can offer for a given context
	///
	/// A basic writer that gives correct results across most contexts should
	/// report `Fallback`, while a specialized writer that really shines in a
	/// given context should report `Supported` when the context is right and
	/// `Unsupported` if the context is not as expected.
	enum class ContextSupport
	{
		Unsupported, //> Material type is not supported
		Supported, ///> Material type is supported
		Fallback //> Material type is not supported, use the fallback (default) writer
	};

	/// A static function is expected for all shader writers and allows
	/// declaring how well this class can support the current context.
	///
	/// The prototype is:
	///
	/// ```static ContextSupport CanExport(const USDSceneBuilderOptions& exportArgs);```
	///

	/// \brief Main export function that runs when the applicable material gets hit.
	MaxUSDAPI virtual void Write() {};

	/// \brief Gets the USD stage that we're writing to.
	MaxUSDAPI const UsdStageRefPtr& GetUsdStage() const;

	/// \brief The path of the destination USD prim to which we are writing.
	MaxUSDAPI const SdfPath& GetUsdPath() const;

	//// \brief The destination USD prim to which we are writing.
	MaxUSDAPI const UsdPrim& GetUsdPrim() const;

	/// \brief The 3ds Max material element being written by this writer.
	MaxUSDAPI Mtl* GetMaterial() const;

	/// \brief The filename to which the WriteJob exports to
	MaxUSDAPI const std::string& GetFilename() const;

protected:
	/// \brief Sets the destination USD prim to which we are writing. (Should only be used once in the
	/// constructor)
	MaxUSDAPI void SetUsdPrim(const UsdPrim& usdPrim);

	/// \brief Gets the current global export args in effect.
	MaxUSDAPI const USDSceneBuilderOptions& GetExportArgs() const;

	UsdPrim usdPrim;
	MaxUsdWriteJobContext& writeJobCtx;
private:
	Mtl* material;
	SdfPath usdPath;
};

typedef std::shared_ptr<MaxUsdShaderWriter> MaxUsdShaderWriterSharedPtr;

PXR_NAMESPACE_CLOSE_SCOPE
