#pragma once
//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
// These coded instructions, statements, and computer programs contain
// unpublished proprietary information written by Autodesk, Inc., and are
// protected by Federal copyright law. They may not be disclosed to third
// parties or copied or duplicated in any form, in whole or in part, without
// the prior written consent of Autodesk, Inc.
//**************************************************************************/

#include "WriteJobContext.h"
#include "ShaderWriter.h"
#include "MaxUsd/Builders/USDSceneBuilderOptions.h"

#include <pxr/pxr.h>
#include <pxr/usd/sdf/path.h>

#include <maxtypes.h>

#include <functional>

PXR_NAMESPACE_OPEN_SCOPE

struct MaxUsdShaderWriterRegistry
{
    // Writer factory function, i.e. a function that creates a shader writer
    // for the given 3ds Max node/USD paths and context.
    using WriterFactoryFn = std::function<MaxUsdShaderWriterSharedPtr(
        Mtl*,
        const SdfPath&,
        MaxUsdWriteJobContext&)>;

    // Predicate function, i.e. a function that can tell the level of support
    // the writer function will provide for a given set of export options.
    using ContextPredicateFn = std::function<MaxUsdShaderWriter::ContextSupport(const USDSceneBuilderOptions&)>;

    // Register 'fn' as a factory function providing a
    // MaxUsdShaderWriter subclass that can be used to write the material ClassID.
    // If you can't provide a valid MaxUsdShaderWriter for the given arguments,
    // return a null pointer from the factory function 'fn'.
    MaxUSDAPI static void Register(
        const TfToken&     maxClassName,
        ContextPredicateFn pred,
        WriterFactoryFn    fn,
        bool               fromPython = false);
    MaxUSDAPI static void Register(
        const Class_ID&    maxClassID,
        ContextPredicateFn pred,
        WriterFactoryFn    fn,
        bool               fromPython = false);

    // Finds a writer if one exists for 3ds Max material ClassID using the context
    // found in 'exportArgs'
    // If there is no writer plugin for 3ds Max material ClassID, returns nullptr.
    MaxUSDAPI static WriterFactoryFn Find(
        const Class_ID& maxTypeName,
        const USDSceneBuilderOptions& exportArgs);
};

/// SFINAE utility class to detect the presence of a CanExport static function
/// inside a writer class. Used by the registration macro for basic writers.
template <typename T> class HasCanExport
{
    typedef char _One;
    struct _Two
    {
        char _x[2];
    };

    template <typename C> static _One _Test(decltype(&C::CanExport));
    template <typename C> static _Two _Test(...);

public:
    enum
    {
        value = sizeof(_Test<T>(0)) == sizeof(char)
    };
};

#define PXR_MAXUSD_REGISTER_SHADER_WRITER(maxClassID, writerClass)                          \
    TF_REGISTRY_FUNCTION_WITH_TAG(MaxUsdShaderWriterRegistry, writerClass)                 \
    {                                                                                      \
        static_assert(std::is_base_of<MaxUsdShaderWriter, writerClass>::value,             \
                #writerClass " must derive from MaxUsdShaderWriter");                      \
        static_assert(                                                                     \
            HasCanExport<writerClass>::value,                                              \
            #writerClass " must define a static CanExport() function");                    \
        MaxUsdShaderWriterRegistry::Register(                                              \
            maxClassID,                                                                    \
            &writerClass::CanExport,                                                       \
            [](Mtl* material,                                                              \
               const SdfPath& usd_path,                                                    \
               MaxUsdWriteJobContext& job_ctx)                                             \
            {                                                                              \
                return std::make_shared<writerClass>(material, usd_path, job_ctx);         \
            });                                                                            \
    }

PXR_NAMESPACE_CLOSE_SCOPE
