#pragma once
//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
// Use of this software is subject to the terms of the Autodesk license 
// agreement provided at the time of installation or download, or which 
// otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/

#include "MaxUsd/MaxUSDAPI.h"
#include <MaxUsd/Translators/ShadingModeExporter.h>

#include <pxr/base/tf/registryManager.h>
#include <pxr/base/tf/singleton.h>
#include <pxr/base/tf/staticTokens.h>
#include <pxr/base/tf/token.h>
#include <pxr/base/tf/weakBase.h>
#include <pxr/pxr.h>

#include <string>

PXR_NAMESPACE_OPEN_SCOPE

// clang-format off
// ShadingMode - shading schema to use
//         none - export no shading data to the USD
//  useRegistry - registry based to export the 3ds Max materials to an equivalent UsdShade network
// additional ShadingMode types can be added thru the use of the RegisterExporter method
#define PXR_MAXUSD_SHADINGMODE_TOKENS \
	(none) \
	(useRegistry)
// clang-format on

TF_DECLARE_PUBLIC_TOKENS(
	MaxUsdShadingModeTokens,
	MaxUSDAPI,
	PXR_MAXUSD_SHADINGMODE_TOKENS);

// clang-format off
// ShadingConversion - shading schema to use
//         none - export no shading data to the USD
//  useRegistry - registry based to export the 3ds Max materials to an equivalent UsdShade network
// additional ShadingMode types can be added thru the use of the RegisterExporter method
#define PXR_MAXUSD_SHADINGCONVERSION_TOKENS \
	(none) \
	(usdPreviewSurface)
// clang-format on

TF_DECLARE_PUBLIC_TOKENS(
	MaxUsdPreferredMaterialTokens,
	MaxUSDAPI,
	PXR_MAXUSD_SHADINGCONVERSION_TOKENS);

TF_DECLARE_WEAK_PTRS(MaxUsdShadingModeRegistry);

/// This class provide macros that are entry points into the shading import/export
/// logic.
///
/// We understand that shading may want to be imported/exported in many ways
/// across studios.  Even within a studio, different workflows may call for
/// different shading modes.

/// The useRegistry exporters and importers can be specialized to support material conversions.
/// The most well known is the default conversion to UsdPreviewSurface shaders. This registry
/// allows introducing other material conversions as necessary to support other renderers. We
/// also allow specifying that an import path is available for these renderers if support has
/// been implemented.
///
/// To register a material conversion on export, you need to use the
/// REGISTER_SHADING_MODE_EXPORT_MATERIAL_CONVERSION macro for each material conversion
/// supported by the library. Multiple registration is supported, so each plugin should declare
/// once the material conversions it supports.
///
/// In order for the core system to discover the plugin, you need a \c plugInfo.json that
/// declares the plugin exposes shading modes:
/// \code
/// {
///   "Plugins": [
///     {
///       "Info": {
///         "MaxUsd": {
///          "ShadingModePlugin" : {}
///         }
///       },
///       "Name": "myUsdPlugin",
///       "LibraryPath": "../myUsdPlugin.dll",
///       "Type": "library"
///     }
///   ]
/// }
/// \endcode
///
/// The plugin at LibraryPath will be loaded via the regular USD plugin loading mechanism.
class MaxUsdShadingModeRegistry : public TfWeakBase
{
public:
	static MaxUsdShadingModeExporterCreator GetExporter(const TfToken& name)
	{
		return GetInstance()._GetExporter(name);
	}
	static TfTokenVector ListExporters() { return GetInstance()._ListExporters(); }

	/// Gets the nice name of an exporter. Used for the UI label of the export options
	static const std::string& GetExporterNiceName(const TfToken& name)
	{
		return GetInstance()._GetExporterNiceName(name);
	}

	/// Gets the description of an exporter. Used for the popup help of the export options
	static const std::string& GetExporterDescription(const TfToken& name)
	{
		return GetInstance()._GetExporterDescription(name);
	}

	MaxUSDAPI
	static MaxUsdShadingModeRegistry& GetInstance();

	MaxUSDAPI
	bool RegisterExporter(
		const std::string&               name,
		std::string                      niceName,
		std::string                      description,
		MaxUsdShadingModeExporterCreator fn);



	/// Get all registered export conversions:
	static TfTokenVector ListMaterialConversions()
	{
		return GetInstance()._ListMaterialConversions();
	}

	/// All the information registered for a specific material conversion.
	struct ConversionInfo
	{
		TfToken renderContext;
		TfToken niceName;
		TfToken exportDescription;
		bool    hasExporter = false;

		ConversionInfo() = default;
		ConversionInfo(TfToken rc, TfToken nn, TfToken ed, bool he)
			: renderContext(rc)
			, niceName(nn)
			, exportDescription(ed)
			, hasExporter(he)
		{
		}
	};

	/// Gets the conversion information associated with \p materialConversion on export and import
	static const ConversionInfo& GetMaterialConversionInfo(const TfToken& materialConversion)
	{
		return GetInstance()._GetMaterialConversionInfo(materialConversion);
	}

	/// Registers an export material conversion with render context, nice name, and description.
	///
	/// The materialConversion name gets used directly in the render option string as one of 
	/// the "Materials export to" options of the USD export dialog.
	///
	/// The renderContext gets used to specialize the binding point. See UsdShadeMaterial documentation
	/// for details. Use a value of "UsdShadeTokens->universalRenderContext" if the resulting UsdShade 
	/// nodes are written using an API shared by multiple renderers, like UsdPreviewSurface. For UsdShade
	/// nodes targeting a specific rendering engine, please define a custom render context
	///  understood by the renderer.
	///
	/// The niceName is the name displayed in the "Materials export to" option of the USD export dialog
	///
	/// The description gets displayed as a tooltip in the "Materials export to" option of the USD export dialog.
	MaxUSDAPI
	void RegisterExportConversion(
		const TfToken& materialConversion,
		const TfToken& renderContext,
		const TfToken& niceName,
		const TfToken& description);

private:
	MaxUSDAPI
	MaxUsdShadingModeExporterCreator _GetExporter(const TfToken& name);
	MaxUSDAPI const std::string& _GetExporterNiceName(const TfToken&);
	MaxUSDAPI const std::string& _GetExporterDescription(const TfToken&);

	MaxUSDAPI TfTokenVector _ListExporters();

	MaxUSDAPI TfTokenVector _ListMaterialConversions();
	MaxUSDAPI const ConversionInfo& _GetMaterialConversionInfo(const TfToken&);

	MaxUsdShadingModeRegistry();
	~MaxUsdShadingModeRegistry();
	friend class TfSingleton<MaxUsdShadingModeRegistry>;
};

#define REGISTER_SHADING_MODE_EXPORT_MATERIAL_CONVERSION(                   \
	name, renderContext, niceName, description)                             \
	TF_REGISTRY_FUNCTION(MaxUsdShadingModeExportContext)                    \
	{                                                                       \
		MaxUsdShadingModeRegistry::GetInstance().RegisterExportConversion(  \
			name, renderContext, niceName, description);                    \
	}

PXR_NAMESPACE_CLOSE_SCOPE
