//**************************************************************************/
// Copyright (c) 1998-2019 Autodesk, Inc.
// All rights reserved.
//
// Use of this software is subject to the terms of the Autodesk license 
// agreement provided at the time of installation or download, or which 
// otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/
// DESCRIPTION: Logging utility.
// AUTHOR: Autodesk Inc.
//***************************************************************************/
#pragma once

#include "spdlog/spdlog.h"
#include "spdlog/common.h"
#include "spdlog/sinks/basic_file_sink.h"

// MAXX-63363: VS2019 v142: <experimental/filesystem> is deprecated and superseded by the C++17 <filesystem> header
#if _MSVC_LANG > 201402L
#include <filesystem>
namespace fs = std::filesystem;
#else
#include <experimental/filesystem>
namespace fs = std::experimental::filesystem;
#endif

#include "MaxUsd/MaxUSDAPI.h"
#include "MaxUsd.h"

#pragma warning(push)
#pragma warning(disable : 4251) // class 'boost::shared_ptr<T>' needs to have dll-interface to be used by clients of class maxUsd::Log

namespace MAXUSD_NS_DEF {

/// Simple wrapper for some basic logging functionality provided by the spdlog library.
class MaxUSDAPI Log {

	/// The current spd logger in use. Can be null.
	static std::shared_ptr<spdlog::logger> spdLogger;

	/// Max file size for logs.
	static const size_t maxLogSize = 1048576 * 200;
	/// How many rotating log files.
	static const size_t maxLogFiles = 5;

	static bool paused;
	
public:

	/**
	 * \brief logging severity level for filtering
	 * \comment value of enum is important as it reflects the index in the ui of qcombobox
	 */
	enum class Level
	{
		Off = 0,
		Error = 1,
		Warn = 2,
		Info = 3
	};

	struct Options
	{
		fs::path path;
		Level level = Level::Off;
	};

	/// The Session class is useful for setting up logging within a scope, and using RAII to destroy the logger.
	/// Starting a logging session will close any previously active sessions.
	class Session
	{
	public:
		Session(const std::string& name, const Options& options);
		~Session();
	};

	static void Message(Level messageType, const std::wstring& message);

	static void Pause();
	static void Resume();

	template<typename FormatString, typename... Args>
	static void Warn(const FormatString &fmt, const Args &... args)
	{
		if (spdLogger && !paused) {
			spdLogger->warn(fmt, args...);
		}
	}

	template<typename FormatString, typename... Args>
	static void Info(const FormatString &fmt, const Args &... args)
	{
		if (spdLogger && !paused) {
			spdLogger->info(fmt, args...);
		}
	}

	template<typename FormatString, typename... Args>
	static void Error(const FormatString &fmt, const Args &... args)
	{
		if (spdLogger  && !paused) {
			spdLogger->error(fmt, args...);
		}
	}
};

}

#pragma warning(pop)
