//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
//  Use of this software is subject to the terms of the Autodesk license 
//  agreement provided at the time of installation or download, or which 
//  otherwise accompanies this software in either electronic or hard copy form.   
//**************************************************************************/

#include "GlTFMaterialWriter.h"
#include "Materials/Mtl.h"
#include <iparamb2.h>
#include <notify.h>
#include <MaxUsd/Translators/ShaderWriterRegistry.h>
#include <units.h>

#include <pxr/usd/usdShade/shader.h>
#include <pxr/usdImaging/usdImaging/tokens.h>

// Macro for the pixar namespace "pxr::"
// This is putting all of the code, until the close macro, into the pixar namespace.
// This is needed for the Macros to compile and is required for all the Prim/Shader Writers
PXR_NAMESPACE_OPEN_SCOPE

// This macro registers the Shader Writer, it's adding the GlTFMaterialWriter class as a candidate when trying to export
// a glTF material. It also verifies that the class we're registering implements the CanExport() method and is a subclass
// of MaxUsdShaderWriter. The class_ID  represents the glTF material.
// It is also very important to set the project option "Remove unreferenced code and data" to NO,
// not doing so could cause the Macro to be optimized out and the Writer to never be properly registered.
PXR_MAXUSD_REGISTER_SHADER_WRITER(Class_ID(0x38420192, 0x45fe4e1b), GlTFMaterialWriter);

PXR_NAMESPACE_CLOSE_SCOPE

/*
 * When this function is called we already know we are dealing with a glTF material,
 * because we register this writer specifically against the glTF material Class_ID (see registration macro above).
 * Which means the only thing we have to check is that we're trying to export to the desired format.
 */
pxr::MaxUsdShaderWriter::ContextSupport GlTFMaterialWriter::CanExport(const USDSceneBuilderOptions& exportArgs)
{
	// GetConvertMaterialsTo returns the current target material being processed
	// In this sample our target is UsdPreviewSurface.
	if (exportArgs.GetConvertMaterialsTo() == pxr::UsdImagingTokens->UsdPreviewSurface)
	{
		return ContextSupport::Supported;
	}
	// Only report as fallback if UsdPreviewSurface was not explicitly requested:
	if (exportArgs.GetAllMaterialConversions().count(pxr::UsdImagingTokens->UsdPreviewSurface) == 0)
	{
		return ContextSupport::Fallback;
	}
	return ContextSupport::Unsupported;
}

/*
 * The Shader Writer constructor is expected to define the Shader prim.
 * The Write method is then responsible for populating it's data.
 */
GlTFMaterialWriter::GlTFMaterialWriter(Mtl* material, const pxr::SdfPath& usdPath, pxr::MaxUsdWriteJobContext& jobCtx)
		: MaxUsdShaderWriter(material, usdPath, jobCtx)
{
	pxr::UsdShadeShader shaderSchema = pxr::UsdShadeShader::Define(GetUsdStage(), GetUsdPath());
	if (!shaderSchema)
	{
		maxUsd::Log::Error(
				"GlTFMaterialWriter : Could not define UsdShadeShader at path '{0}'\n", GetUsdPath().GetString());
		return;
	}

	pxr::UsdAttribute idAttr = shaderSchema.CreateIdAttr(pxr::VtValue(pxr::UsdImagingTokens->UsdPreviewSurface));

	usdPrim = shaderSchema.GetPrim();
	if (!usdPrim.IsValid())
	{
		maxUsd::Log::Error("GlTFMaterialWriter : Could not get UsdPrim for UsdShadeShader at path '{0}'\n",
				shaderSchema.GetPath().GetString());
		return;
	}

	// Surface Output
	shaderSchema.CreateOutput(pxr::UsdShadeTokens->surface, pxr::SdfValueTypeNames->Token);
}

/*
 * For the purpose of this sample, which is demonstrating the necessary parts needed to implement a Shader Writer, we
 * will just export the base color of the glTF material. A similar approach can be taken for the other parameters.
 */
void GlTFMaterialWriter::Write()
{
	pxr::UsdShadeShader shaderSchema(usdPrim);
	if (!shaderSchema)
	{
		maxUsd::Log::Error("GlTFMaterialWriter : Could not get UsdShadeShader schema for UsdPrim at path '{0}'\n",
				usdPrim.GetPath().GetString());
		return;
	}
	const auto timeConfig = GetExportArgs().GetResolvedTimeConfig();
	Mtl* mat = GetMaterial();
	Interval valid = FOREVER;
	auto pb = mat->GetParamBlock(0);
	Point3 col;
	const TimeValue startTime = maxUsd::getTicksFromFrame(timeConfig.startFrame);
	const TimeValue endTime = maxUsd::getTicksFromFrame(timeConfig.endFrame);
	// How much time we need to put for each Max's frame.
	const int timeStep = int(double(GetTicksPerFrame()) / double(timeConfig.samplesPerFrame));
	pxr::UsdShadeInput in = shaderSchema.CreateInput(pxr::TfToken("diffuseColor"), pxr::SdfValueTypeNames->Color3f);

	for (TimeValue timeVal = startTime; timeVal <= endTime;)
	{
		pb->GetValueByName(_T("baseColor"), timeVal, col, valid);
		pxr::UsdTimeCode usdTimeCode(double(timeVal) / double(GetTicksPerFrame()));
		in.Set(pxr::GfVec3f(col[0], col[1], col[2]), usdTimeCode);
		timeVal += timeStep;
	}
}
