//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
//  Use of this software is subject to the terms of the Autodesk license 
//  agreement provided at the time of installation or download, or which 
//  otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/

#include "SpherePrimWriter.h"
#include <iparamb2.h>
#include <notify.h>
#include <object.h>

#include <MaxUsd/MeshConversion/MeshConverter.h>
#include <MaxUsd/Utilities/TranslationUtils.h>
#include <MaxUsd/Translators/PrimWriterRegistry.h>

#include <pxr/usd/usdShade/shader.h>
#include <pxr/usd/usdGeom/sphere.h>

// Macro for the pixar namespace "pxr::"
// This is putting all of the code, until the close macro, into the pixar namespace.
// This is needed for the Macros to compile and is required for all the Prim/Shader Writers
PXR_NAMESPACE_OPEN_SCOPE

// This macro registers the Prim Writer, it's adding the SpherePrimWriter class as a candidate when trying to export
// an object. Unlike for shader writer, this doesn't contain information about the supported classes.
// The CanConvert() method is responsible for defining what can be exported or not.
// It is also really important to set the project option "Remove unreferenced code and data" to NO,
// this could cause the Macro to be optimized out and the Writer to never be properly registered.
PXR_MAXUSD_REGISTER_WRITER(SpherePrimWriter);

PXR_NAMESPACE_CLOSE_SCOPE

// For demonstration purposes, provide an example to export using a native usd sphere, or as a usd mesh.
static bool EXPORT_AS_NATIVE_SPHERE = true;

/*
 * This method is responsible of telling the export process if it can export the current node's object.
 * In the case of this sample, the only object type we want to handle is the Sphere.
 */
pxr::MaxUsdPrimWriter::ContextSupport SpherePrimWriter::CanConvert(INode* node)
{
	// We simply pass TimeValue "0" here to EvalWorldStage() as it is not important at what time we evaluate the object,
	// we are only looking at the object's type and we assume that it will not change over time.
	const auto object = node->EvalWorldState(0).obj;
	if (object->ClassID() == Class_ID(SPHERE_CLASS_ID, 0))
	{
		return ContextSupport::Supported;
	}
	return ContextSupport::Unsupported;
}

/*
 * This Writer only deals with Sphere objects - we return a "Sphere" token if we want to convert the node to a native
 * USD Sphere, or mesh, if not. For performance reasons, the export is done in two passes. The first pass creates all
 * the prims inside a single SdfChangeBlock, the second pass populate each prim's attributes. It is not mandatory to
 * implement this function, it is also possible to define the prim from the Write() method (which would override the
 * prim's type created in the first pass), but you would lose the performance benefit. If not implemented, the base
 * implementation returns Xform.
 */
pxr::TfToken SpherePrimWriter::GetPrimType(INode* node)
{
	return EXPORT_AS_NATIVE_SPHERE ? pxr::TfToken("Sphere") : pxr::TfToken("Mesh");
}

SpherePrimWriter::SpherePrimWriter(const pxr::MaxUsdWriteJobContext& jobCtx)
		: MaxUsdPrimWriter(jobCtx)
{
}

/*
 * For this sample, we will demonstrate how to export the Radius and DisplayColor attributes.
 * We'll also demonstrate one way to handle animation.
 */
bool SpherePrimWriter::Write(INode* sourceNode, pxr::UsdPrim& targetPrim, bool applyOffsetTransform)
{
	if (EXPORT_AS_NATIVE_SPHERE)
	{
		// targetPrim is already a Sphere, it was created for us from the type returned in GetPrimType()
		pxr::UsdGeomSphere spherePrim(targetPrim);
		pxr::UsdAttribute radiusAttr = spherePrim.CreateRadiusAttr();
		pxr::UsdAttribute extentAttr = spherePrim.CreateExtentAttr();
		const auto spherePb = sourceNode->EvalWorldState(0).obj->GetParamBlock(0);
		float radius = 0;
		Interval inter = FOREVER;

		const auto timeConfig = GetExportArgs().GetResolvedTimeConfig();
		// getFramesFromValidityInterval will return all the frames we need to export to have an accurate
		// representations of the scene and of the chosen options in the exporter. It computes the frames based on the
		// validity interval of the object.
		const auto frames = maxUsd::getFramesFromValidityInterval(sourceNode, timeConfig);
		for (size_t i = 0; i < frames.first.size(); i++)
		{
			// Get the value in Max time.
			// We already know we're dealing with a Sphere object at this point, because we wouldn't be here otherwise.
			// CanConvert must have passed for Write to be called.
			spherePb->GetValueByName(_T("Radius"), frames.first[i], radius, inter);
			// Set it at the equivalent usd time.
			// When doing set on Attribute, the type must match exactly, otherwise it'll do nothing.
			radiusAttr.Set(static_cast<double>(radius), frames.second[i]);
			// Set the extent Attribute, this is usd to compute the bounding box in USD
			pxr::VtVec3fArray extent = { pxr::GfVec3f(-radius, -radius, -radius), 
				pxr::GfVec3f(radius, radius, radius) };
			extentAttr.Set(extent, frames.second[i]);
		}

		pxr::UsdAttribute displayColorAttr = spherePrim.CreateDisplayColorAttr();
		const COLORREF wireCol = sourceNode->GetWireColor();
		pxr::VtVec3fArray colVec;
		colVec.push_back(pxr::GfVec3f(static_cast<int>(GetRValue(wireCol)) / 255.f,
				static_cast<int>(GetGValue(wireCol)) / 255.f, static_cast<int>(GetBValue(wireCol)) / 255.f));
		// When doing set on Attribute, the type must match exactly, otherwise it'll do nothing.
		displayColorAttr.Set(colVec);
	}
	// Alternatively, we could export the Sphere as a mesh, using the MeshConverter utility :
	else
	{
		MeshConverter converter;
		converter.ConvertToUSDMesh(sourceNode, targetPrim.GetStage(), targetPrim.GetPath(),
				GetExportArgs().GetMeshConversionOptions(), GetExportArgs().GetResolvedTimeConfig(),
				applyOffsetTransform);
	}
	return true;
}