import maxUsd
from pymxs import runtime as rt
from pxr import UsdGeom
from pxr import Gf as pyGf
import pymxs
import traceback

EXPORT_AS_NATIVE_SPHERE = True

# Prim writer for exporting Sphere object to USD native sphere.
class SphereWriter(maxUsd.PrimWriter):

    # This Writer only deals with Sphere objects, so we return a "Sphere" token because we want to convert the node to a native USD Sphere.
    # For performance reasons, the export is done in two passes.
    # The first pass creates all the prims inside a single SdfChangeBlock, the second pass populate each prim's attributes.
    # It is not mandatory to implement this function, it is also possible to define the prim from the Write() method (which would override the prim's type created in the first pass),
    # but you would lose the performance benefit. If not implemented, the base implementation returns Xform.
    def GetPrimType(self, nodeHandle):
        if EXPORT_AS_NATIVE_SPHERE:
            return "Sphere"
        return "Mesh"

    # For this sample, we will demonstrate how to export the Radius and DisplayColor attributes.
    # We'll also demonstrate one way to handle animation.    
    def Write(self, nodeHandle, prim, applyOffset):
        try: 
            stage = prim.GetStage()
            opts = self.GetExportArgs()
            node = rt.maxOps.getNodeByHandle(nodeHandle)
            if EXPORT_AS_NATIVE_SPHERE:            
                # prim is already a Sphere, it was created for us from the type returned in GetPrimType()
                spherePrim = UsdGeom.Sphere(prim)
                radiusAttr = spherePrim.GetRadiusAttr()
                extentAttr = spherePrim.CreateExtentAttr()
                # getFramesFromValidityInterval will return all the frames we need to export to have an accurate representations of the scene and of the chose option in the export.
                # It computes the frames based on the validity interval of the object.
                frames = maxUsd.TranslationUtils.GetKeyFramesFromValidityInterval(nodeHandle, opts)
                for maxFrame,usdTime in frames:
                    with pymxs.attime(maxFrame):
                        radiusAttr.Set(node.radius, usdTime)
                        extent = pyGf.Vec3f(node.radius, node.radius, node.radius)
                        extentAttr.Set([-extent, extent], usdTime)
                print("Write a Sphere as a USD Sphere prim!")
            # Alternatively, we could export the Sphere as a mesh, using the MeshConverter utility :
            else:
                spherePrim = maxUsd.MeshConverter.ConvertToUSDMesh(nodeHandle, stage, prim.GetPath(), opts, applyOffset)
                print("Write a Sphere as a USD Mesh prim!")
                
            # Setup the wireColor as displayColor.
            displayColorAttr = spherePrim.GetDisplayColorAttr()
            r = node.wireColor.r / 255
            g = node.wireColor.g / 255
            b = node.wireColor.b / 255
            displayColorAttr.Set([(r,g,b)])
            
            return True
        except Exception as e:
            # Quite useful to debug errors in a Python callback
            print('Write() - Error: %s' % str(e))
            print(traceback.format_exc())
            return False

    # This method is responsible of telling the export process if it can export the current node's object.
    # In the case of this sample, the only object type we want to handle is the Sphere.    
    def CanConvert(self, nodeHandle):
        node = rt.maxOps.getNodeByHandle(nodeHandle)
        if rt.classOf(node) == rt.Sphere:
            print("Found a sphere...")
            return maxUsd.PrimWriter.ContextSupport.Supported
        return maxUsd.PrimWriter.ContextSupport.Unsupported
   
# Register the writer.
# First argument is the class, second argument is the Writer name, which will be used as an ID internaly.
maxUsd.PrimWriter.Register(SphereWriter, "SphereWriter")