//**************************************************************************/
// Copyright (c) 2022 Autodesk, Inc.
// All rights reserved.
//
//  Use of this software is subject to the terms of the Autodesk license 
//  agreement provided at the time of installation or download, or which 
//  otherwise accompanies this software in either electronic or hard copy form.  
//**************************************************************************/

#include <MaxUsd/Chaser/ExportChaser.h>
#include <MaxUsd/Chaser/ExportChaserRegistry.h>
#include <MaxUsd/Builders/SceneBuilderOptions.h>
#include <MaxUsd/Builders/USDSceneBuilderOptions.h>
#include <MaxUsd/Builders/JobContextRegistry.h>

#include <object.h>
#include <icustattribcontainer.h>
#include <custattrib.h>
#include <maxscript/mxsplugin/mxsCustomAttributes.h>

#include <boost/tokenizer.hpp>

// Data types suported by the UserDataExportChaserSample
enum class PropertyType: int
{
	USER_PROP = 1,   // "user"
	CUSTOM_DATA = 2  // "custom"
};

// The UserDataExportChaserSample class
// This Export Chaser sample adds Custom Data to the exported prims if the specified user
// property or custom attribute are found in the corresponding Node. The sample chaser
// acts with default data or using the arguments that may have been passed to the chaser.
class UserDataExportChaserSample : public pxr::MaxUsdExportChaser
{
public:
	// constructor - can be customized for the Export Chaser requirements.
	// In the provided sample it receives the stage and Prims to Node map, which are 
	// the minimal arguments for a functional chaser, and the chaser arguments, which are
	// not mandatory but can be used to parameterize the chaser.
	UserDataExportChaserSample(
		pxr::UsdStagePtr stage,
		const pxr::MaxUsdExportChaserRegistry::FactoryContext::PrimsToNodesMap& primsToNodesMap,
		const USDSceneBuilderOptions::ChaserArgs& args);

	// processing that needs to run after the main 3ds Max USD export loop.
	bool PostExport() override;

private:
	// the exported stage
	pxr::UsdStagePtr stage;
	// maps full USD prim paths to INodes
	const pxr::MaxUsdExportChaserRegistry::FactoryContext::PrimsToNodesMap& primsToNodesMap;

	// sample specific parameters
	// maps the data to be exported as Custom Data for the prims
	typedef std::map<PropertyType, std::set<std::string>> PropertyMap;
	PropertyMap dataToExport;

};

// Add the using statement here to make the export chaser factory macro work
PXR_NAMESPACE_USING_DIRECTIVE
// Macro to register the Export Chaser. Defines a factory method for the chaser name. The 'ctx' will
// be of type MaxUsdExportChaserRegistry::FactoryContext. The method should return a MaxUsdExportChaser*.
// There are no guarantees about the lifetime of 'ctx'.
// It is also very important to set the project option "Remove unreferenced code and data" to NO,
// not doing so could cause the Macro to be optimized out and the Export Chaser to never be properly registered.
PXR_MAXUSD_DEFINE_EXPORT_CHASER_FACTORY(UserData, "User Data C++ DEMO", "Chaser to export user data along the exported USD prims", ctx)
{
	// fetching the export chaser parameters
	// the chaser arguments are mapped using their registration chaser name (first arg of the macro)
	USDSceneBuilderOptions::ChaserArgs myArgs;
	TfMapLookup(ctx.GetJobArgs().GetAllChaserArgs(), "UserData", &myArgs);

	return new UserDataExportChaserSample(ctx.GetStage(), ctx.GetPrimsToNodesMap(), myArgs);
}

UserDataExportChaserSample::UserDataExportChaserSample(
	pxr::UsdStagePtr stage,
	const pxr::MaxUsdExportChaserRegistry::FactoryContext::PrimsToNodesMap& primsToNodesMap,
	const USDSceneBuilderOptions::ChaserArgs& args)
	: stage(stage)
	, primsToNodesMap(primsToNodesMap)
{
	// default configuration
	dataToExport = { { PropertyType::USER_PROP, { "myUserProperty" } }, { PropertyType::CUSTOM_DATA, {} } };

	// parsing the specific export chaser arguments
	boost::char_separator<char> sep(",");
	for (std::pair<std::string, std::string> item : args)
	{
		if (item.first == "user")
		{
			dataToExport[PropertyType::USER_PROP].clear();
			boost::tokenizer<boost::char_separator<char>> tokens(item.second, sep);
			for (auto name : tokens)
			{
				dataToExport[PropertyType::USER_PROP].insert(name);
			}
		}
		else if (item.first == "custom")
		{
			dataToExport[PropertyType::CUSTOM_DATA].clear();
			boost::tokenizer<boost::char_separator<char>> tokens(item.second, sep);
			for (auto name : tokens)
			{
				dataToExport[PropertyType::CUSTOM_DATA].insert(name);
			}
		}
		else
		{
			maxUsd::Log::Warn(
					"Wrong user data type ('{0}') passed as argument to UserPropertyExportChaser\n", item.first);
		}
	}
}

bool UserDataExportChaserSample::PostExport()
{
	// cycle through the exported prim/node
 	for (std::pair<std::string, INode*> exported : primsToNodesMap)
	{
		pxr::UsdPrim prim = stage->GetPrimAtPath(pxr::SdfPath(exported.first));
		INode* node = exported.second;

		// are any of the specified user properties found in the Node
		for (std::string propName : dataToExport[PropertyType::USER_PROP])
		{
			auto mxsPropName = maxUsd::usdStringToMaxString(propName);
			if (node->UserPropExists(mxsPropName))
			{
				TSTR value;
				node->GetUserPropString(mxsPropName, value);

				if (_tcsicmp(value, _T("true")) == 0)
				{
					prim.SetCustomDataByKey(pxr::TfToken(propName), pxr::VtValue(true));
				}
				else if (_tcsicmp(value, _T("false")) == 0)
				{
					prim.SetCustomDataByKey(pxr::TfToken(propName), pxr::VtValue(false));
				}
				else
				{
					// note - to simplify the chaser sample implementation, the user property conversion
					// to numerical values was not completed. It will be written as a simple string. The
					// same chaser written in Python properly handles this situation however.
					auto valueString = maxUsd::maxStringToUSDString(value);
					// strip the quoting marks from the string (if the property was originally a string)
					if (*valueString.begin() == '\"' && *(valueString.end() - 1) == '\"')
					{
						valueString.erase(valueString.begin());
						valueString.erase(valueString.end()-1);
					}
					prim.SetCustomDataByKey(pxr::TfToken(propName), pxr::VtValue(valueString));
				}
			}
		}

		// for the purpose of the sample, the chaser finds the custom attributes only on
		// the base level, not on the modifiers or materials of the node
		BaseObject* obj = node->GetObjectRef();
		ICustAttribContainer* cc = obj->GetCustAttribContainer();
		if (!cc)
		{
			continue;
		}
		// are any of the specified custom attributes found in the Node
		for (std::string propName : dataToExport[PropertyType::CUSTOM_DATA])
		{
			auto mxsPropName = maxUsd::usdStringToMaxString(propName);
			for (int i = 0; i < cc->GetNumCustAttribs(); i++)
			{
				CustAttrib* ca = cc->GetCustAttrib(i);
				if (!ca)
				{
					continue;
				}
				MSCustAttrib* msCA = (MSCustAttrib*)ca->GetInterface(I_SCRIPTEDCUSTATTRIB);
				if (!msCA)
				{
					continue;
				}

				for (int j = 0; j < ca->NumParamBlocks(); ++j)
				{
					IParamBlock2* pb2 = ca->GetParamBlock(j);
					if (!pb2)
					{
						continue;
					}
					ParamID paramId = maxUsd::FindParamId(pb2, mxsPropName);
					// If this specific attribute doesn't exist, we skip to look for it in the next PB
					if (paramId == -1)
					{
						continue;
					}

					// found the attribute, extract the value and write the data
					{
						const ParamDef& paramDef = pb2->GetParamDef(paramId);
						Interval iv = FOREVER;
						bool hasFoundParam = false;
						switch (paramDef.type)
						{
						case TYPE_BOOL: {
							int boolVal = 0;
							hasFoundParam = pb2->GetValue(paramId, 0, boolVal, iv);
							prim.SetCustomDataByKey(
									pxr::TfToken(propName), pxr::VtValue((boolVal == 0) ? false : true));
							break;
						}
						case TYPE_INT: {
							int strVal = false;
							hasFoundParam = pb2->GetValue(paramId, 0, strVal, iv);
							prim.SetCustomDataByKey(pxr::TfToken(propName), pxr::VtValue(strVal));
							break;
						}
						case TYPE_FLOAT: {
							float strVal = false;
							hasFoundParam = pb2->GetValue(paramId, 0, strVal, iv);
							prim.SetCustomDataByKey(pxr::TfToken(propName), pxr::VtValue(strVal));
							break;
						}
						case TYPE_STRING: {
							const wchar_t* strVal = L"";
							hasFoundParam = pb2->GetValue(paramId, 0, strVal, iv);
							prim.SetCustomDataByKey(
									pxr::TfToken(propName), pxr::VtValue(maxUsd::maxStringToUSDString(strVal)));
							break;
						}
						default:
							maxUsd::Log::Warn(
									"Unsupported Custom Attribute type for '{0}' in UserPropertyExportChaser\n",
									propName);
							break;
						}
						// no need to search further, the attribute was found
						break;
					}
				}
			}
		}
	}
	return true;
}

REGISTER_EXPORT_JOB_CONTEXT_FCT(CustomDemoContext, "Custom Context C++ DEMO", "Custom plug-in configuration")
{
	VtDictionary extraArgs;
	extraArgs[MaxUsdUsdSceneBuilderOptionsTokens->chaser] = VtValue(std::vector<VtValue> { VtValue(std::string("UserData")) } );
	VtValue chaserArgUserProp(std::vector<VtValue> {
									VtValue(std::string("UserData")),
									VtValue(std::string("user")),
									VtValue(std::string("myUserFloatProperty,myUserProperty")) });
	VtValue chaserArgCustomProp(std::vector<VtValue> {
									VtValue(std::string("UserData")),
									VtValue(std::string("custom")),
									VtValue(std::string("inGame")) });
	extraArgs[MaxUsdUsdSceneBuilderOptionsTokens->chaserArgs] = VtValue(std::vector<VtValue>{ chaserArgUserProp, chaserArgCustomProp });
	
	return extraArgs;
}